@@ -20,6 +20,8 @@ typedef struct {
 #define EXPORT_PEM 0
 #define EXPORT_DER 1
 
+static const rb_data_type_t ossl_ec_group_type;
+static const rb_data_type_t ossl_ec_point_type;
 
 #define GetPKeyEC(obj, pkey) do { \
     GetPKey((obj), (pkey)); \
@@ -30,7 +32,7 @@ typedef struct {
 
 #define SafeGet_ec_group(obj, group) do { \
     OSSL_Check_Kind((obj), cEC_GROUP); \
-    Data_Get_Struct((obj), ossl_ec_group, (group)); \
+    TypedData_Get_Struct((obj), ossl_ec_group, &ossl_ec_group_type, (group)); \
 } while(0)
 
 #define Get_EC_KEY(obj, key) do { \
@@ -52,7 +54,7 @@ typedef struct {
 
 #define Get_EC_GROUP(obj, g) do { \
     ossl_ec_group *ec_group; \
-    Data_Get_Struct((obj), ossl_ec_group, ec_group); \
+    TypedData_Get_Struct((obj), ossl_ec_group, &ossl_ec_group_type, ec_group); \
     if (ec_group == NULL) \
         ossl_raise(eEC_GROUP, "missing ossl_ec_group structure"); \
     (g) = ec_group->group; \
@@ -71,7 +73,7 @@ typedef struct {
 
 #define Get_EC_POINT(obj, p) do { \
     ossl_ec_point *ec_point; \
-    Data_Get_Struct((obj), ossl_ec_point, ec_point); \
+    TypedData_Get_Struct((obj), ossl_ec_point, &ossl_ec_point_type, ec_point); \
     if (ec_point == NULL) \
         ossl_raise(eEC_POINT, "missing ossl_ec_point structure"); \
     (p) = ec_point->point; \
@@ -114,6 +116,7 @@ static VALUE ec_instance(VALUE klass, EC_KEY *ec)
     if (!ec) {
 	return Qfalse;
     }
+    obj = NewPKey(klass);
     if (!(pkey = EVP_PKEY_new())) {
 	return Qfalse;
     }
@@ -121,7 +124,7 @@ static VALUE ec_instance(VALUE klass, EC_KEY *ec)
 	EVP_PKEY_free(pkey);
 	return Qfalse;
     }
-    WrapPKey(klass, obj, pkey);
+    SetPKey(obj, pkey);
 
     return obj;
 }
@@ -133,10 +136,11 @@ VALUE ossl_ec_new(EVP_PKEY *pkey)
     if (!pkey) {
 	obj = ec_instance(cEC, EC_KEY_new());
     } else {
+	obj = NewPKey(cEC);
 	if (EVP_PKEY_type(pkey->type) != EVP_PKEY_EC) {
 	    ossl_raise(rb_eTypeError, "Not a EC key!");
 	}
-	WrapPKey(cEC, obj, pkey);
+	SetPKey(obj, pkey);
     }
     if (obj == Qfalse) {
 	ossl_raise(eECError, NULL);
@@ -369,7 +373,7 @@ static VALUE ossl_ec_point_dup(const EC_POINT *point, VALUE group_v)
     ossl_ec_point *new_point;
 
     obj = rb_obj_alloc(cEC_POINT);
-    Data_Get_Struct(obj, ossl_ec_point, new_point);
+    TypedData_Get_Struct(obj, ossl_ec_point, &ossl_ec_point_type, new_point);
 
     SafeRequire_EC_GROUP(group_v, group);
 
@@ -493,7 +497,10 @@ static VALUE ossl_ec_key_to_string(VALUE self, VALUE ciph, VALUE pass, int forma
 	    if (!NIL_P(ciph)) {
 		cipher = GetCipherPtr(ciph);
 		if (!NIL_P(pass)) {
-		    password = StringValuePtr(pass);
+		    StringValue(pass);
+		    if (RSTRING_LENINT(pass) < OSSL_MIN_PWD_LEN)
+			ossl_raise(eOSSLError, "OpenSSL requires passwords to be at least four characters long");
+		    password = RSTRING_PTR(pass);
 		}
 	    }
 	    else {
@@ -530,8 +537,8 @@ static VALUE ossl_ec_key_to_string(VALUE self, VALUE ciph, VALUE pass, int forma
 
 /*
  *  call-seq:
- *     key.to_pem   => String
- *     key.to_pem(cipher, pass_phrase) => String
+ *     key.export([cipher, pass_phrase]) => String
+ *     key.to_pem([cipher, pass_phrase]) => String
  *
  * Outputs the EC key in PEM encoding.  If +cipher+ and +pass_phrase+ are
  * given they will be used to encrypt the key.  +cipher+ must be an
@@ -540,7 +547,7 @@ static VALUE ossl_ec_key_to_string(VALUE self, VALUE ciph, VALUE pass, int forma
  * text.
  *
  */
-static VALUE ossl_ec_key_to_pem(int argc, VALUE *argv, VALUE self)
+static VALUE ossl_ec_key_export(int argc, VALUE *argv, VALUE self)
 {
     VALUE cipher, passwd;
     rb_scan_args(argc, argv, "02", &cipher, &passwd);
@@ -700,21 +707,32 @@ static VALUE ossl_ec_key_dsa_verify_asn1(VALUE self, VALUE data, VALUE sig)
     }
 
     ossl_raise(eECError, "ECDSA_verify");
+
+    UNREACHABLE;
 }
 
-static void ossl_ec_group_free(ossl_ec_group *ec_group)
+static void ossl_ec_group_free(void *ptr)
 {
+    ossl_ec_group *ec_group = ptr;
     if (!ec_group->dont_free && ec_group->group)
         EC_GROUP_clear_free(ec_group->group);
     ruby_xfree(ec_group);
 }
 
+static const rb_data_type_t ossl_ec_group_type = {
+    "OpenSSL/ec_group",
+    {
+	0, ossl_ec_group_free,
+    },
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
 static VALUE ossl_ec_group_alloc(VALUE klass)
 {
     ossl_ec_group *ec_group;
     VALUE obj;
 
-    obj = Data_Make_Struct(klass, ossl_ec_group, 0, ossl_ec_group_free, ec_group);
+    obj = TypedData_Make_Struct(klass, ossl_ec_group, &ossl_ec_group_type, ec_group);
 
     return obj;
 }
@@ -741,7 +759,7 @@ static VALUE ossl_ec_group_initialize(int argc, VALUE *argv, VALUE self)
     ossl_ec_group *ec_group;
     EC_GROUP *group = NULL;
 
-    Data_Get_Struct(self, ossl_ec_group, ec_group);
+    TypedData_Get_Struct(self, ossl_ec_group, &ossl_ec_group_type, ec_group);
     if (ec_group->group != NULL)
         ossl_raise(rb_eRuntimeError, "EC_GROUP is already initialized");
 
@@ -842,6 +860,7 @@ static VALUE ossl_ec_group_initialize(int argc, VALUE *argv, VALUE self)
 }
 
 /*  call-seq:
+ *     group1.eql?(group2)   => true | false
  *     group1 == group2   => true | false
  *
  */
@@ -1213,19 +1232,28 @@ static VALUE ossl_ec_group_to_text(VALUE self)
 }
 
 
-static void ossl_ec_point_free(ossl_ec_point *ec_point)
+static void ossl_ec_point_free(void *ptr)
 {
+    ossl_ec_point *ec_point = ptr;
     if (!ec_point->dont_free && ec_point->point)
         EC_POINT_clear_free(ec_point->point);
     ruby_xfree(ec_point);
 }
 
+static const rb_data_type_t ossl_ec_point_type = {
+    "OpenSSL/ec_point",
+    {
+	0, ossl_ec_point_free,
+    },
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
 static VALUE ossl_ec_point_alloc(VALUE klass)
 {
     ossl_ec_point *ec_point;
     VALUE obj;
 
-    obj = Data_Make_Struct(klass, ossl_ec_point, 0, ossl_ec_point_free, ec_point);
+    obj = TypedData_Make_Struct(klass, ossl_ec_point, &ossl_ec_point_type, ec_point);
 
     return obj;
 }
@@ -1246,7 +1274,7 @@ static VALUE ossl_ec_point_initialize(int argc, VALUE *argv, VALUE self)
     VALUE group_v = Qnil;
     const EC_GROUP *group = NULL;
 
-    Data_Get_Struct(self, ossl_ec_point, ec_point);
+    TypedData_Get_Struct(self, ossl_ec_point, &ossl_ec_point_type, ec_point);
     if (ec_point->point)
         ossl_raise(eEC_POINT, "EC_POINT already initialized");
 
@@ -1311,6 +1339,7 @@ static VALUE ossl_ec_point_initialize(int argc, VALUE *argv, VALUE self)
 
 /*
  *  call-seq:
+ *     point1.eql?(point2) => true | false
  *     point1 == point2 => true | false
  *
  */
@@ -1353,6 +1382,8 @@ static VALUE ossl_ec_point_is_at_infinity(VALUE self)
     case 0: return Qfalse;
     default: ossl_raise(cEC_POINT, "EC_POINT_is_at_infinity");
     }
+
+    UNREACHABLE;
 }
 
 /*
@@ -1374,6 +1405,8 @@ static VALUE ossl_ec_point_is_on_curve(VALUE self)
     case 0: return Qfalse;
     default: ossl_raise(cEC_POINT, "EC_POINT_is_on_curve");
     }
+
+    UNREACHABLE;
 }
 
 /*
@@ -1465,6 +1498,78 @@ static VALUE ossl_ec_point_to_bn(VALUE self)
     return bn_obj;
 }
 
+/*
+ *  call-seq:
+ *     point.mul(bn)  => point
+ *     point.mul(bn, bn) => point
+ *     point.mul([bn], [point]) => point
+ *     point.mul([bn], [point], bn) => point
+ */
+static VALUE ossl_ec_point_mul(int argc, VALUE *argv, VALUE self)
+{
+    EC_POINT *point1, *point2;
+    const EC_GROUP *group;
+    VALUE group_v = rb_iv_get(self, "@group");
+    VALUE bn_v1, bn_v2, r, points_v;
+    BIGNUM *bn1 = NULL, *bn2 = NULL;
+
+    Require_EC_POINT(self, point1);
+    SafeRequire_EC_GROUP(group_v, group);
+
+    r = rb_obj_alloc(cEC_POINT);
+    ossl_ec_point_initialize(1, &group_v, r);
+    Require_EC_POINT(r, point2);
+
+    argc = rb_scan_args(argc, argv, "12", &bn_v1, &points_v, &bn_v2);
+
+    if (rb_obj_is_kind_of(bn_v1, cBN)) {
+        bn1 = GetBNPtr(bn_v1);
+        if (argc >= 2) {
+            bn2 = GetBNPtr(points_v);
+        }
+        if (EC_POINT_mul(group, point2, bn2, point1, bn1, ossl_bn_ctx) != 1)
+            ossl_raise(eEC_POINT, "Multiplication failed");
+    } else {
+        size_t i, points_len, bignums_len;
+        const EC_POINT **points;
+        const BIGNUM **bignums;
+
+        Check_Type(bn_v1, T_ARRAY);
+        bignums_len = RARRAY_LEN(bn_v1);
+        bignums = (const BIGNUM **)OPENSSL_malloc(bignums_len * (int)sizeof(BIGNUM *));
+
+        for (i = 0; i < bignums_len; ++i) {
+            bignums[i] = GetBNPtr(rb_ary_entry(bn_v1, i));
+        }
+
+        if (!rb_obj_is_kind_of(points_v, rb_cArray)) {
+            OPENSSL_free((void *)bignums);
+            rb_raise(rb_eTypeError, "Argument2 must be an array");
+        }
+
+        rb_ary_unshift(points_v, self);
+        points_len = RARRAY_LEN(points_v);
+        points = (const EC_POINT **)OPENSSL_malloc(points_len * (int)sizeof(EC_POINT *));
+
+        for (i = 0; i < points_len; ++i) {
+            Get_EC_POINT(rb_ary_entry(points_v, i), points[i]);
+        }
+
+        if (argc >= 3) {
+            bn2 = GetBNPtr(bn_v2);
+        }
+        if (EC_POINTs_mul(group, point2, bn2, points_len, points, bignums, ossl_bn_ctx) != 1) {
+            OPENSSL_free((void *)bignums);
+            OPENSSL_free((void *)points);
+            ossl_raise(eEC_POINT, "Multiplication failed");
+        }
+        OPENSSL_free((void *)bignums);
+        OPENSSL_free((void *)points);
+    }
+
+    return r;
+}
+
 static void no_copy(VALUE klass)
 {
     rb_undef_method(klass, "copy");
@@ -1473,7 +1578,7 @@ static void no_copy(VALUE klass)
     rb_undef_method(klass, "initialize_copy");
 }
 
-void Init_ossl_ec()
+void Init_ossl_ec(void)
 {
 #ifdef DONT_NEED_RDOC_WORKAROUND
     mOSSL = rb_define_module("OpenSSL");
@@ -1531,7 +1636,8 @@ void Init_ossl_ec()
     rb_define_method(cEC, "dsa_verify_asn1", ossl_ec_key_dsa_verify_asn1, 2);
 /* do_sign/do_verify */
 
-    rb_define_method(cEC, "to_pem", ossl_ec_key_to_pem, -1);
+    rb_define_method(cEC, "export", ossl_ec_key_export, -1);
+    rb_define_alias(cEC, "to_pem", "export");
     rb_define_method(cEC, "to_der", ossl_ec_key_to_der, 0);
     rb_define_method(cEC, "to_text", ossl_ec_key_to_text, 0);
 
@@ -1585,6 +1691,7 @@ void Init_ossl_ec()
 /* all the other methods */
 
     rb_define_method(cEC_POINT, "to_bn", ossl_ec_point_to_bn, 0);
+    rb_define_method(cEC_POINT, "mul", ossl_ec_point_mul, -1);
 
     no_copy(cEC);
     no_copy(cEC_GROUP);
@@ -1592,7 +1699,7 @@ void Init_ossl_ec()
 }
 
 #else /* defined NO_EC */
-void Init_ossl_ec()
+void Init_ossl_ec(void)
 {
 }
 #endif /* NO_EC */
