@@ -1,7 +1,8 @@
+# frozen_string_literal: true
 #
 # tmpdir - retrieve temporary directory path
 #
-# $Id: tmpdir.rb 63017 2018-03-28 14:34:14Z usa $
+# $Id: tmpdir.rb 52526 2015-11-10 11:48:14Z akr $
 #
 
 require 'fileutils'
@@ -19,7 +20,7 @@ class Dir
 
   def self.tmpdir
     if $SAFE > 0
-      @@systmpdir
+      @@systmpdir.dup
     else
       tmp = nil
       [ENV['TMPDIR'], ENV['TMP'], ENV['TEMP'], @@systmpdir, '/tmp', '.'].each do |dir|
@@ -108,12 +109,10 @@ class Dir
     def make_tmpname((prefix, suffix), n)
       prefix = (String.try_convert(prefix) or
                 raise ArgumentError, "unexpected prefix: #{prefix.inspect}")
-      prefix = prefix.delete("#{File::SEPARATOR}#{File::ALT_SEPARATOR}")
       suffix &&= (String.try_convert(suffix) or
                   raise ArgumentError, "unexpected suffix: #{suffix.inspect}")
-      suffix &&= suffix.delete("#{File::SEPARATOR}#{File::ALT_SEPARATOR}")
       t = Time.now.strftime("%Y%m%d")
-      path = "#{prefix}#{t}-#{$$}-#{rand(0x100000000).to_s(36)}"
+      path = "#{prefix}#{t}-#{$$}-#{rand(0x100000000).to_s(36)}".dup
       path << "-#{n}" if n
       path << suffix if suffix
       path
